Object.defineProperty(exports, '__esModule', { value: true });

var path = require('path');
var fs = require('fs');

const {
  access: accessAsync,
  readFile: readAsync,
  stat: statAsync
} = fs.promises;
const has = Function.call.bind(hasOwnProperty);
/**
 * Returns an object from JavaScript Object Notation (JSON).
 * @arg {string} json
 * @return {Object<string, any>}
 */

function parseJson(json) {
  try {
    return Object(JSON.parse(json));
  } catch (error) {
    return {};
  }
}
/**
 * Synchronously returns the contents of a file, otherwise null.
 * @arg {string} file
 * @arg {ReadFileSyncOptions} options
 * @return {string | void}
 */

function readSync(file, options) {
  try {
    return fs.readFileSync(file, options);
  } catch (error) {
    return null;
  }
}
/**
 * Synchronously returns information about a file.
 * @arg {string} file
 * @arg {StatFileSystemOptions} options
 * @return {import('fs').BigIntStats | void}
 */

function statSync(file, mode) {
  try {
    return fs.statSync(file, mode);
  } catch (error) {
    return null;
  }
}
/**
 * Return whether the id starts with a root path separator.
 * @param {string} id
 * @param {Options} opts
 */

function startsWithRoot(id, opts) {
  return RegExp('^\\' + opts.pathSeparator).test(id);
}
/**
 * Return whether the id starts with a relative path separator.
 * @param {string} id
 * @param {Options} opts
 */

function startsWithRelative(id, opts) {
  return RegExp('^\\.{0,2}\\' + opts.pathSeparator).test(id);
}
/**
 * Return a list of possible "node_modules" directories.
 * @arg {string} cwd
 * @arg {Options} opts
 */

function nodeModulesDirs(cwd, opts) {
  // segments is `cwd` split by `/`
  const segments = cwd.split(opts.pathSeparator); // `count` is the length of segments

  let count = segments.length; // `dirs` is an empty list

  const dirs = []; // while `count` is greater than `0`

  while (count > 0) {
    // if `segments[count]` is not `node_modules`
    if (segments[count] !== 'node_modules') {
      // push a new path to `dirs` as the `/`-joined `segments[0 - count]` and `node_modules`
      dirs.push(path.join(segments.slice(0, count).join(opts.pathSeparator) || opts.pathSeparator, 'node_modules'));
    } // `count` is `count` minus `1`


    --count;
  } // return `dirs`


  return dirs;
}
/** @typedef {{ encoding: BufferEncoding; flag?: string; } | BufferEncoding} ReadFileSyncOptions */

/** @typedef {{ bigint: boolean }} StatFileSystemOptions */

/** @typedef {string | void} SettledPathSync */

/** @typedef {Promise<string | Error>} SettledPathAsync */

/** @typedef {{ baseUrl: string, extensions: string[], fileResolverAsync: (file: string, opts: Options) => SettledPathAsync, fileResolverAsyncCache: Object<string, SettledPathAsync>, fileResolverSync: (file: string, opts: Options) => SettledPathSync, fileResolverSyncCache: Object<string, SettledPathSync>, indexes: string[], packageProps: string[], packageResolverAsync: (pkg: Object<string, any>, opts: Options) => Promise<string | void>, packageResolverSync: (pkg: Object<string, any>, opts: Options) => string | void, pathSeparator: string }} Options */

/**
 * Return a resolved path object if a file exists, otherwise a rejection.
 * @arg {string} file
 * @arg {Options} opts
 * @return {SettledPathAsync}
 */

function fileStatAsync(file, opts) {
  if (!has(opts.fileResolverAsyncCache, file)) {
    opts.fileResolverAsyncCache[file] = statAsync(file).then(stat => stat.isFile() ? file : Promise.reject());
  }

  return opts.fileResolverAsyncCache[file];
}
/**
 * Return a path object if a file exists, otherwise null.
 * @arg {string} file
 * @arg {Options} opts
 * @return {SettledPathSync}
 */

function fileStatSync(file, opts) {
  if (!has(opts.fileResolverSyncCache, file)) {
    const stat = statSync(file);
    opts.fileResolverSyncCache[file] = stat && stat.isFile() ? file : null;
  }

  return opts.fileResolverSyncCache[file];
}
/** @typedef {string | void} SettledPathSync */

/** @typedef {Promise<string | Error>} SettledPathAsync */

/** @typedef {{ baseUrl: string, extensions: string[], fileResolverAsync: (file: string, opts: Options) => SettledPathAsync, fileResolverAsyncCache: Object<string, SettledPathAsync>, fileResolverSync: (file: string, opts: Options) => SettledPathSync, fileResolverSyncCache: Object<string, SettledPathSync>, indexes: string[], packageProps: string[], packageResolverAsync: (pkg: Object<string, any>, opts: Options) => Promise<string | void>, packageResolverSync: (pkg: Object<string, any>, opts: Options) => string | void, pathSeparator: string }} Options */

/**
 * Resolve the package value, otherwise rejection.
 * @arg {Object<string, any>} pkg
 * @return {Promise<string | void>}
 */
function packageResolverAsync(pkg, opts) {
  const value = packageResolverSync(pkg, opts);
  return value ? Promise.resolve(value) : Promise.reject();
}
/**
 * Return the package value, otherwise null.
 * @arg {Object<string, any>} pkg
 * @return {string | void}
 */

function packageResolverSync(pkg, opts) {
  const value = opts.packageProps.reduce((packageValue, packageProp) => packageValue || packageProp.split('.').reduce((object, field) => Object(object)[field], pkg), null);
  return typeof value === 'string' ? value : null;
}

/**
 * Return a resolved file object if a file exists, otherwise a rejection.
 * @arg {string} file
 * @arg {Options} opts
 * @return {SettledPathAsync}
 */
function resolveAsFileAsync(file, opts) {
  return (// resolve `file` as the file
    opts.fileResolverAsync(file, opts).catch( // otherwise, for each `fileExtension` of `extensions`
    () => opts.extensions.reduce((fileInfoAsync, fileExtension) => fileInfoAsync.catch( // resolve `file.fileExtension` as the file
    () => opts.fileResolverAsync(`${file}.${fileExtension}`, opts)), Promise.reject()))
  );
}
/**
 * Return a file object if a file exists, otherwise null.
 * @arg {string} file
 * @arg {Options} opts
 * @return {SettledPathSync}
 */

function resolveAsFileSync(file, opts) {
  return (// resolve `file` as the file
    opts.fileResolverSync(file, opts) || // otherwise, for each `fileExtension` of `extensions`
    opts.extensions.reduce((fileInfoSync, fileExtension) => fileInfoSync || // resolve `file.fileExtension` as the file
    opts.fileResolverSync(`${file}.${fileExtension}`, opts), null)
  );
}
/** @typedef {string | void} SettledPathSync */

/** @typedef {Promise<string | Error>} SettledPathAsync */

/** @typedef {{ baseUrl: string, extensions: string[], fileResolverAsync: (file: string, opts: Options) => SettledPathAsync, fileResolverAsyncCache: Object<string, SettledPathAsync>, fileResolverSync: (file: string, opts: Options) => SettledPathSync, fileResolverSyncCache: Object<string, SettledPathSync>, indexes: string[], packageProps: string[], packageResolverAsync: (pkg: Object<string, any>, opts: Options) => Promise<string | void>, packageResolverSync: (pkg: Object<string, any>, opts: Options) => string | void, pathSeparator: string }} Options */

/**
 * Return a resolved file path if a file exists, otherwise a rejection.
 * @arg {string} dir
 * @arg {Options} opts
 * @return {SettledPathAsync}
 */

function resolveAsDirectoryAsync(dir, opts) {
  // if package.json exists
  return readAsync(path.join(dir, 'package.json'), 'utf8').then(contents => {
    // resolve the JSON contents of `dir/package.json` as `pkg`
    const pkg = parseJson(contents); // if `pkg` has the resolved file field

    return opts.packageResolverAsync(pkg, opts).then( // resolve the resolved file field as the file
    file => opts.fileResolverAsync(path.join(dir, file), opts), // otherwise, resolve by file index
    resolveByindexAsync);
  }, // otherwise, resolve by file index
  resolveByindexAsync);

  function resolveByindexAsync() {
    // for each `index` of `indexes`
    return opts.indexes.reduce((fileInfoAsync, index) => fileInfoAsync.catch( // resolve `dir/index` as the file
    () => opts.fileResolverAsync(path.join(dir, index), opts)), Promise.reject());
  }
}
/**
 * Return a file path if a file exists, otherwise null.
 * @arg {string} dir
 * @arg {Options} opts
 * @return {SettledPathSync}
 */

function resolveAsDirectorySync(dir, opts) {
  const contents = readSync(path.join(dir, 'package.json'), 'utf8'); // if package.json exists

  if (contents) {
    // resolve the JSON contents of `dir/package.json` as `pkg`
    const pkg = parseJson(contents);
    const file = opts.packageResolverSync(pkg, opts); // if `pkg` has the resolved file field

    if (file) {
      // resolve the resolved file field as the file
      return opts.fileResolverSync(path.join(dir, file), opts);
    }
  } // otherwise, resolve by file index


  return resolveByIndexSync();

  function resolveByIndexSync() {
    // for each `index` of `indexes`
    return opts.indexes.reduce((fileInfoSync, index) => fileInfoSync || // resolve `dir/index` as the file
    opts.fileResolverSync(path.join(dir, index), opts), null);
  }
}
/** @typedef {string | void} SettledPathSync */

/** @typedef {Promise<string | Error>} SettledPathAsync */

/** @typedef {{ baseUrl: string, extensions: string[], fileResolverAsync: (file: string, opts: Options) => SettledPathAsync, fileResolverAsyncCache: Object<string, SettledPathAsync>, fileResolverSync: (file: string, opts: Options) => SettledPathSync, fileResolverSyncCache: Object<string, SettledPathSync>, indexes: string[], packageProps: string[], packageResolverAsync: (pkg: Object<string, any>, opts: Options) => Promise<string | void>, packageResolverSync: (pkg: Object<string, any>, opts: Options) => string | void, pathSeparator: string }} Options */

/**
* Resolve a module.
* @arg {string} cwd
* @arg {string} id
* @arg {Options} opts
* @return {SettledPathAsync}
*/

function resolveAsModuleAsync(cwd, id, opts) {
  // for each `dir` in the node modules directory using `cwd`
  return nodeModulesDirs(cwd, opts).reduce((fileInfoAsync, dir) => fileInfoAsync.catch( // resolve as a file using `dir/id` as `file`
  () => resolveAsFileAsync(path.join(dir, id), opts) // otherwise, resolve as a directory using `dir/id` as `dir`
  .catch(() => resolveAsDirectoryAsync(path.join(dir, id), opts))), Promise.reject());
}
/**
* Resolve a module.
* @arg {string} cwd
* @arg {string} id
* @arg {Options} opts
* @return {SettledPathSync}
*/

function resolveAsModuleSync(cwd, id, opts) {
  // for each `dir` in the node modules directory using `cwd`
  return nodeModulesDirs(cwd, opts).reduce((fileInfoSync, dir) => fileInfoSync || // resolve as a file using `dir/id` as `file`
  resolveAsFileSync(path.join(dir, id), opts) || // otherwise, resolve as a directory using `dir/id` as `dir`
  resolveAsDirectorySync(path.join(dir, id), opts), null);
}
/** @typedef {string | void} SettledPathSync */

/** @typedef {Promise<string | Error>} SettledPathAsync */

/** @typedef {{ baseUrl: string, extensions: string[], fileResolverAsync: (file: string, opts: Options) => SettledPathAsync, fileResolverAsyncCache: Object<string, SettledPathAsync>, fileResolverSync: (file: string, opts: Options) => SettledPathSync, fileResolverSyncCache: Object<string, SettledPathSync>, indexes: string[], packageProps: string[], packageResolverAsync: (pkg: Object<string, any>, opts: Options) => Promise<string | void>, packageResolverSync: (pkg: Object<string, any>, opts: Options) => string | void, pathSeparator: string }} Options */

const defaultextensions = ['css'];
const defaultIndexes = ['index.css'];
const defaultPackageProps = ['exports.css.import', 'exports.css.default', 'exports.css', 'style'];
const defaultPathSeparator = '/';
/**
 * Resolve the location of a file within `url(id)` from `cwd`
 * @arg {string} id
 * @arg {string} cwd
 * @arg {Partial<Options>} init
 * @return {SettledPathAsync}
 */

function resolveAsync(id, cwd, init) {
  /** @type {Options} */
  const opts = init === Object(init) ? init : {};
  if (!has(opts, 'baseUrl')) opts.baseUrl = cwd;
  if (!has(opts, 'extensions')) opts.extensions = defaultextensions;
  if (!has(opts, 'indexes')) opts.indexes = defaultIndexes;
  if (!has(opts, 'fileResolverAsync')) opts.fileResolverAsync = fileStatAsync;
  if (!has(opts, 'fileResolverAsyncCache')) opts.fileResolverAsyncCache = Object.create(null);
  if (!has(opts, 'fileResolverSync')) opts.fileResolverSync = fileStatSync;
  if (!has(opts, 'fileResolverSyncCache')) opts.fileResolverSyncCache = Object.create(null);
  if (!has(opts, 'packageProps')) opts.packageProps = defaultPackageProps;
  if (!has(opts, 'packageResolverAsync')) opts.packageResolverAsync = packageResolverAsync;
  if (!has(opts, 'packageResolverSync')) opts.packageResolverSync = packageResolverSync;
  if (!has(opts, 'pathSeparator')) opts.pathSeparator = defaultPathSeparator; // if `id` starts with `/`

  if (startsWithRoot(id, opts)) {
    // `cwd` is the filesystem root
    cwd = '';
  }

  return (// resolve as a file using `cwd/id` as `file`
    resolveAsFileAsync(path.join(cwd, id), opts) // otherwise, resolve as a directory using `cwd/id` as `dir`
    .catch(() => resolveAsDirectoryAsync(path.join(cwd, id), opts)).catch(() => {
      // otherwise, if `id` does not begin with `/`, `./`, or `../`
      if (!startsWithRelative(id, opts)) {
        if (opts.baseUrl !== cwd) {
          return resolveAsFileAsync(path.join(opts.baseUrl, id), opts).catch(() => resolveAsDirectoryAsync(path.join(opts.baseUrl, id), opts));
        }

        return resolveAsModuleAsync(cwd, id, opts);
      } // otherwise, return null


      return Promise.reject();
    }).catch( // otherwise, throw `"CSS Module not found"`
    () => Promise.reject(new Error('CSS Module not found')))
  );
}
/**
 * Return the location of a file within `url(id)` from `cwd`
 * @arg {string} id
 * @arg {string} cwd
 * @arg {Partial<Options>} init
 * @return {SettledPathSync}
 */

function resolveSync(id, cwd, init) {
  /** @type {Options} */
  const opts = init === Object(init) ? init : {};
  if (!has(opts, 'baseUrl')) opts.baseUrl = cwd;
  if (!has(opts, 'extensions')) opts.extensions = defaultextensions;
  if (!has(opts, 'indexes')) opts.indexes = defaultIndexes;
  if (!has(opts, 'fileResolverAsync')) opts.fileResolverAsync = fileStatAsync;
  if (!has(opts, 'fileResolverAsyncCache')) opts.fileResolverAsyncCache = Object.create(null);
  if (!has(opts, 'fileResolverSync')) opts.fileResolverSync = fileStatSync;
  if (!has(opts, 'fileResolverSyncCache')) opts.fileResolverSyncCache = Object.create(null);
  if (!has(opts, 'packageProps')) opts.packageProps = defaultPackageProps;
  if (!has(opts, 'packageResolverAsync')) opts.packageResolverAsync = packageResolverAsync;
  if (!has(opts, 'packageResolverSync')) opts.packageResolverSync = packageResolverSync;
  if (!has(opts, 'pathSeparator')) opts.pathSeparator = defaultPathSeparator; // if `id` starts with `/`

  if (startsWithRoot(id, opts)) {
    // `cwd` is the filesystem root
    cwd = '';
  }

  return (// return as a file using `cwd/id` as `file`
    resolveAsFileSync(path.join(cwd, id), opts) || // otherwise, return as a directory using `cwd/id` as `dir`
    resolveAsDirectorySync(path.join(cwd, id), opts) || ( // otherwise, if `id` does not begin with `/`, `./`, or `../`
    !startsWithRelative(id, opts) ? opts.baseUrl !== cwd ? resolveAsFileSync(path.join(opts.baseUrl, id), opts) || resolveAsDirectorySync(path.join(opts.baseUrl, id), opts) : resolveAsModuleSync(cwd, id, opts) : null) || // otherwise, return null
    null
  );
}
/** @typedef {string | void} SettledPathSync */

/** @typedef {Promise<string | Error>} SettledPathAsync */

/** @typedef {{ baseUrl: string, extensions: string[], fileResolverAsync: (file: string, opts: Options) => SettledPathAsync, fileResolverAsyncCache: Object<string, SettledPathAsync>, fileResolverSync: (file: string, opts: Options) => SettledPathSync, fileResolverSyncCache: Object<string, SettledPathSync>, indexes: string[], packageProps: string[], packageResolverAsync: (pkg: Object<string, any>, opts: Options) => Promise<string | void>, packageResolverSync: (pkg: Object<string, any>, opts: Options) => string | void, pathSeparator: string }} Options */

const resolve = {
  async: resolveAsync,
  sync: resolveSync
};

exports.default = resolve;
exports.resolveAsync = resolveAsync;
exports.resolveSync = resolveSync;
